import os
from dataclasses import dataclass
from itertools import chain
from pathlib import Path
from typing import Iterable, Iterator, Sequence

from . import raw

__all__ = ["Document", "Page", "Text"]

OBJ_ID_PAGETREE = 2
OBJ_ID_RESOURCES = 3
OBJ_ID_FIRST_PAGE = 4
A4_SIZE_IN_PT = (595, 842)
DEFAULT_FONT_NAME = "F1"
DEFAULT_FONT_SIZE = 14
MAX_STRING_LENGTH = 65535

PageIndex = int
"""Zero-indexed page number"""


@dataclass(frozen=True)
class Point:
    x: float
    y: float

    def __iter__(self) -> Iterator[float]:
        yield self.x
        yield self.y


@dataclass(frozen=True, init=False)
class Text:
    content: str
    at: Point

    def __init__(self, content: str, at: Point | tuple[float, float]):
        object.__setattr__(self, "content", content)
        object.__setattr__(self, "at", Point(*at))

    def __post_init__(self) -> None:
        try:
            self.content.encode("ascii")
        except UnicodeEncodeError:
            raise NotImplementedError("only ASCII for now")

        assert len(self.content) < MAX_STRING_LENGTH, "text too long!"

    def to_stream(self) -> bytes:
        return f"""\
  BT
    /{DEFAULT_FONT_NAME} {DEFAULT_FONT_SIZE} Tf
    {self.at.x} {self.at.y} Td
    ({self.content}) Tj
  ET
""".encode(
            "ascii"
        )


@dataclass(frozen=True)
class Page:
    content: Sequence[Text] = ()

    def raw_objects(
        self, index: PageIndex
    ) -> Iterable[tuple[raw.ObjectID, raw.Object]]:
        return [
            (index, self.raw_metadata(index + 1)),
            (index + 1, self.raw_content()),
        ]

    def raw_metadata(self, content: raw.ObjectID, /) -> raw.Dictionary:
        return raw.Dictionary(
            {
                "Type": raw.Name("Page"),
                "Parent": raw.Ref(OBJ_ID_PAGETREE),
                "Contents": raw.Ref(content),
                "Resources": raw.Ref(OBJ_ID_RESOURCES),
            },
        )

    def raw_content(self) -> raw.Stream:
        content = b"\n".join(item.to_stream() for item in self.content)
        return raw.Stream({"Length": raw.Int(len(content) - 1)}, content)


def id_for_page(i: PageIndex) -> raw.ObjectID:
    # For now, we represent pages with two objects:
    # the metadata and the content.
    # Therefore, object ID is enumerated twice as fast as page number.
    return (i * 2) + OBJ_ID_FIRST_PAGE


@dataclass(frozen=True)
class Document:
    pages: Sequence[Page]

    def __post_init__(self) -> None:
        assert self.pages, "at least one page required"

    def to_path(self, p: os.PathLike, /) -> None:
        page_ids = range(OBJ_ID_FIRST_PAGE, id_for_page(len(self.pages)), 2)
        headers: list[tuple[raw.ObjectID, raw.Object]] = [
            (
                raw.OBJ_ID_CATALOG,
                raw.Dictionary(
                    {
                        "Type": raw.Name("Catalog"),
                        "Pages": raw.Ref(OBJ_ID_PAGETREE),
                    }
                ),
            ),
            (
                OBJ_ID_PAGETREE,
                raw.Dictionary(
                    {
                        "Type": raw.Name("Pages"),
                        "Kids": raw.Array(
                            tuple(
                                map(
                                    raw.Ref,
                                    page_ids,
                                )
                            )
                        ),
                        "Count": raw.Int(len(self.pages)),
                        "MediaBox": raw.Array(
                            tuple(map(raw.Int, [0, 0, *A4_SIZE_IN_PT]))
                        ),
                    }
                ),
            ),
            (
                OBJ_ID_RESOURCES,
                raw.Dictionary(
                    {
                        "Font": raw.Dictionary(
                            {
                                DEFAULT_FONT_NAME: raw.Dictionary(
                                    {
                                        "Type": raw.Name("Font"),
                                        "Subtype": raw.Name("Type1"),
                                        "BaseFont": raw.Name("Times-Roman"),
                                    }
                                ),
                            }
                        ),
                    }
                ),
            ),
        ]
        Path(p).write_bytes(
            raw.write(
                chain(
                    headers,
                    chain.from_iterable(
                        page.raw_objects(i)
                        for i, page in zip(page_ids, self.pages)
                    ),
                )
            )
        )



"""Low-level PDF objects and operations"""
import abc
import math
import string
from dataclasses import dataclass
from itertools import accumulate, starmap
from typing import Iterable, Mapping, Sequence, Tuple

ALLOWED_NAME_CHARS = string.ascii_letters + "-_*?$" + string.digits
MAX_INT = 2_147_483_647
CHARS_REQUIRING_ESCAPE = "\n\t\r\f\b\\()"
HEADER = b"%PDF-1.3\n%\xc2\xa5\xc2\xb1\xc3\xab\n"
FIRST_OFFSET = len(HEADER)
OBJ_ID_XREF = 0
OBJ_ID_CATALOG = 1
FIRST_XREF_ENTRY = b"0000000000 65535 f\n"


class Object(abc.ABC):
    @abc.abstractmethod
    def write(self) -> bytes:
        ...


ObjectID = int
"""ID of an object in the PDF file. Always >=0"""


@dataclass(frozen=True)
class Bool(Object):
    value: bool

    def write(self) -> bytes:
        return b"true" if self.value else b"false"


@dataclass(frozen=True)
class Name(Object):
    value: str

    def __post_init__(self) -> None:
        assert all(
            map(ALLOWED_NAME_CHARS.__contains__, self.value)
        ), "only simple alphanumeric names for now"

    def write(self) -> bytes:
        return f"/{self.value}".encode()


@dataclass(frozen=True)
class Int(Object):
    value: int

    def __post_init__(self) -> None:
        # TODO: check if this is a real limitation
        assert MAX_INT > self.value > -MAX_INT, "this int is too large"

    def write(self) -> bytes:
        return str(self.value).encode()


@dataclass(frozen=True)
class Float(Object):
    value: float

    def __post_init__(self) -> None:
        assert math.isfinite(self.value), "NaN and Inf not supported"

    def write(self) -> bytes:
        # TODO: check if there's any Python-specific oddities with float repr
        return str(self.value).encode()


@dataclass(frozen=True)
class String(Object):
    value: str

    def write(self) -> bytes:
        assert not any(
            map(self.value.__contains__, CHARS_REQUIRING_ESCAPE)
        ), "parentheses and nonprintable newlines/tabs currently not supported"
        return f"({self.value})".encode()


@dataclass(frozen=True)
class Array(Object):
    items: Tuple[Object, ...]

    def write(self) -> bytes:
        return b"[%s]" % b" ".join(i.write() for i in self.items)


@dataclass(frozen=True)
class Dictionary(Object):
    content: Mapping[str, Object]

    def __post_init__(self) -> None:
        assert len(self.content) <= 4096, "PDF dictionary too large"

    def write(self) -> bytes:
        return _write_dict(self.content)


def _write_dict(d: Mapping[str, Object]) -> bytes:
    return b"<<%s>>" % b" ".join(
        b"/%s %s" % (k.encode(), v.write()) for k, v in d.items()
    )


@dataclass(frozen=True)
class Stream(Object):
    meta: Mapping[str, Object]
    content: bytes

    def __post_init__(self) -> None:
        assert len(self.meta) <= 4096, "PDF dictionary too large"

    def write(self) -> bytes:
        return _write_dict(self.meta) + b"\nstream\n%sendstream" % self.content


@dataclass(frozen=True)
class Ref(Object):
    target: ObjectID

    def write(self) -> bytes:
        return b"%i 0 R" % self.target


def _write_obj(i: ObjectID, o: Object) -> bytes:
    return b"%i 0 obj\n%s\nendobj\n" % (i, o.write())


def write(objs: Iterable[tuple[ObjectID, Object]], /) -> bytes:
    return b"".join(write_iter(objs))


def _write_trailer(
    obj_offsets: Sequence[int], xref_offset: int
) -> Iterable[bytes]:
    yield b"xref\n%i %i\n" % (OBJ_ID_XREF, len(obj_offsets) + 1)
    yield FIRST_XREF_ENTRY
    for offset in obj_offsets:
        yield b"%010i 00000 n\n" % offset
    yield b"trailer\n"
    yield Dictionary(
        {
            "Root": Ref(OBJ_ID_CATALOG),
            "Size": Int(len(obj_offsets)),
        }
    ).write()
    yield b"\nstartxref\n%i\n%%%%EOF" % (xref_offset)


def write_iter(objs: Iterable[tuple[ObjectID, Object]], /) -> Iterable[bytes]:
    yield HEADER
    offsets = [FIRST_OFFSET]
    for chunk in starmap(_write_obj, objs):
        offsets.append(len(chunk))
        yield chunk

    *obj_offsets, xref_offset = accumulate(offsets)
    yield from _write_trailer(obj_offsets, xref_offset)



# -*- coding: utf-8 -*-
from setuptools import setup

package_dir = \
{'': 'src'}

packages = \
['pdfje']

package_data = \
{'': ['*']}

extras_require = \
{':python_version < "3.8"': ['importlib-metadata>=1,<5']}

setup_kwargs = {
    'name': 'pdfje',
    'version': '0.1.0',
    'description': 'Tiny PDF writer',
    'long_description': '🖍 PDFje\n========\n\n.. image:: https://img.shields.io/pypi/v/pdfje.svg?style=flat-square\n   :target: https://pypi.python.org/pypi/pdfje\n\n.. image:: https://img.shields.io/pypi/l/pdfje.svg?style=flat-square\n   :target: https://pypi.python.org/pypi/pdfje\n\n.. image:: https://img.shields.io/pypi/pyversions/pdfje.svg?style=flat-square\n   :target: https://pypi.python.org/pypi/pdfje\n\n.. image:: https://img.shields.io/readthedocs/pdfje.svg?style=flat-square\n   :target: http://pdfje.readthedocs.io/\n\n.. image:: https://img.shields.io/badge/code%20style-black-000000.svg?style=flat-square\n   :target: https://github.com/psf/black\n\n-----\n\n  **PDF·je** *(noun)* Dutch for \'small PDF\'\n\nTiny library for writing simple PDFs. Experimental.\n\nCurrently under development.\nLeave a ⭐️ on GitHub if you\'re interested how this develops!\n\nWhy?\n----\n\nThe most popular libraries for writing PDFs are quite old and inspired by Java and PHP.\n*PDFje* aims to be a modern, Pythonic library with a more declarative API.\n\nHow does it work?\n-----------------\n\n.. code-block:: python\n\n  >>> from pdfje import Document, Page, Text\n  >>> pdf.Document([\n  ...     Page([Text("Hello", at=(200, 700)), Text("World", at=(300, 670))]),\n  ...     Page(),\n  ...     Page([Text("This is the last page!", at=(300, 600))]),\n  ...\n  ... ]).to_path(\'hello.pdf\')\n\nSee `the docs <https://pdfje.rtfd.io>`_ for a complete overview.\n\nInstallation\n------------\n\nIt\'s available on PyPI.\n\n.. code-block:: bash\n\n  pip install pdfje\n',
    'author': 'Arie Bovenberg',
    'author_email': 'a.c.bovenberg@gmail.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/ariebovenberg/pdfje',
    'package_dir': package_dir,
    'packages': packages,
    'package_data': package_data,
    'extras_require': extras_require,
    'python_requires': '>=3.7,<4.0',
}


setup(**setup_kwargs)
